package uprrp.tania;

import android.app.Application;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Build;
import android.util.Log;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresApi;

import com.amazonaws.auth.CognitoCachingCredentialsProvider;
import com.amazonaws.regions.Regions;
import com.google.android.gms.tasks.OnFailureListener;
import com.google.android.gms.tasks.OnSuccessListener;
import com.google.firebase.iid.FirebaseInstanceId;
import com.google.firebase.iid.InstanceIdResult;

import org.researchstack.backbone.StorageAccess;
import org.researchstack.backbone.storage.database.AppDatabase;
import org.researchstack.backbone.storage.database.sqlite.DatabaseHelper;
import org.researchstack.backbone.storage.file.EncryptionProvider;
import org.researchstack.backbone.storage.file.FileAccess;
import org.researchstack.backbone.storage.file.PinCodeConfig;
import org.researchstack.backbone.storage.file.SimpleFileAccess;
import org.researchstack.backbone.storage.file.UnencryptedProvider;

import uprrp.tania.activities.GettingStartedActivity;

public class App extends Application {

    private static final String TAG = "App";

    @Override
    public void onCreate() {
        super.onCreate();

        // Customize your pin code preferences
        PinCodeConfig pinCodeConfig = new PinCodeConfig(); // default pin config (4-digit, 1 min lockout)

        // Customize encryption preferences
        EncryptionProvider encryptionProvider = new UnencryptedProvider(); // No pin, no encryption

        // If you have special file handling needs, implement FileAccess
        FileAccess fileAccess = new SimpleFileAccess();

        // If you have your own custom database, implement AppDatabase
        AppDatabase database = new DatabaseHelper(this,
                DatabaseHelper.DEFAULT_NAME,
                null,
                DatabaseHelper.DEFAULT_VERSION);

        StorageAccess.getInstance().init(pinCodeConfig, encryptionProvider, fileAccess, database);

        // Get device token
        FirebaseInstanceId.getInstance().getInstanceId()
                .addOnSuccessListener(new OnSuccessListener<InstanceIdResult>() {
                    @Override
                    public void onSuccess(InstanceIdResult instanceIdResult) {

                        // Store token globally
                        String token = instanceIdResult.getToken();
                        GlobalValues.getInstance().setDeviceToken(token);
                        Log.d(TAG, "Got token: " + GlobalValues.getInstance().getDeviceToken());

                        // Enable notifications (NEEDS TOKEN TO BE GLOBALLY AVAILABLE)
                        CognitoCachingCredentialsProvider credentialsProvider = new CognitoCachingCredentialsProvider(
                                getApplicationContext(),
                                getString(R.string.identityPoolID), // Identity pool ID
                                Regions.US_EAST_1 // Region
                        );
                        SNSRegister snsRegister = new SNSRegister(credentialsProvider);
                        snsRegister.execute();

                    }
                })
                .addOnFailureListener(new OnFailureListener() {
                    @Override
                    public void onFailure(@NonNull Exception e) {
                        GlobalValues.getInstance().setDeviceToken(null);
                        Toast.makeText(getApplicationContext(), "Couldn't retrieve token!", Toast.LENGTH_LONG).show();
                        Log.e(TAG, "Error occurred while retrieving token...");
                        e.printStackTrace();
                    }
                });

        // If user hasn't already registered, redirect accordingly,
        // JUST IN CASE user starts non-default activity from somewhere else (like a browser)
        // TODO: find a way to not redirect twice to GettingStarted
        SharedPreferences prefs = getSharedPreferences("prefs", MODE_PRIVATE);
        boolean needsToRegister = prefs.getBoolean("needsToRegister", true);
        if(needsToRegister) {
            Log.wtf(TAG, "Redirecting to GettingStarted");
            Intent intent = new Intent(this, GettingStartedActivity.class);
            intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
            startActivity(intent);
        }

    }
}
